// paTan - silhouette puzzle editor for web browser
// copyright kys, 2020

var applicationVersion='v0.0'

$(window).on('load', onLoad);

//------ application instance
class Patan{
	constructor(){
		this._storage = null;
		this._storageKey = 'paTanSilhouettePuzzle';
		this._localize = new Localize();
		this._colorTheme = new ColorTheme();
		this._jte = null;
		this._listFilter = null;
		this._silhouettePuzzle = {};
		this._twsPieceSetList = [];
		this._width = 0;
		this._height = 0;
	}

	get localize(){ return this._localize; }
	get colorTheme(){ return this._colorTheme; }
	get jte(){ return this._jte; }
	get pieceset(){ return this._silhouettePuzzle.pieceset; }
	get silhouettes(){ return this._silhouettePuzzle.silhouettes; }
	get silhouettesList(){ return this._listFilter.silhouettes; }
	get twsPieceSetList(){ return this._twsPieceSetList; }
	set twsPieceSetList(value){ this._twsPieceSetList = value; }
	get width(){ return this._width; }
	set width(value){ this._width = value; }
	get height(){ return this._height; }
	set height(value){ this._height = value; }

	// initialize
	initialize(scope){
		// init storage and jtelib
		this._storage = sessionStorage;
		this._jte = new Jte(scope);
		this._listFilter = new ListFilter(this, scope);
		if( !this.restoreSilhouettePuzzle() ){
			this._jte.expandInitialJteXml(this._silhouettePuzzle);
		}
		this._listFilter.reset();
	}

	// list
	getRealSilhouetteIndex(listIndex){
		return this._listFilter.getRealSilhouetteIndex(listIndex);
	}

	addSilhouette(silhouette){
		this._silhouettePuzzle.silhouettes.push(silhouette);
		this._listFilter.refresh();
		this.backupSilhouettePuzzle();
	}

	insertSilhouette(index, silhouette){
		this._silhouettePuzzle.silhouettes.splice(index, 0, silhouette);
		this._listFilter.refresh();
		this.backupSilhouettePuzzle();
	}

	setSilhouette(index, silhouette){
		this._silhouettePuzzle.silhouettes[index] = silhouette;
		this._listFilter.refresh();
		this.backupSilhouettePuzzle();
	}

	removeSilhouette(index){
		this._silhouettePuzzle.silhouettes.splice(index, 1);
		this._listFilter.refresh();
		this.backupSilhouettePuzzle();
	}

	setPiecesetSilhouette(silhouette){
		this._silhouettePuzzle.pieceset.silhouette.p = silhouette.p;
		this._silhouettePuzzle.pieceset.silhouette.name = silhouette.name;
		this.backupSilhouettePuzzle();
	}

	setTagFilterByIndex(index){
		this._listFilter.setTagFilterByIndex(index);
	}

	getSilhouetteTagsHtml(classInfo){
		return this._listFilter.getSilhouetteTagsHtml(classInfo);
	}

	// sort
	sortByFeatures(){
		var patan = this;
		var features = [];
		patan.silhouettes.forEach(function(silhouette){
			if( !('feature' in silhouette) ){
				silhouette.feature = patan.jte.getFeatureVector(patan.pieceset, silhouette);
			}
			features.push(silhouette.feature);
		});
		var gauge = patan.jte.buildFeatureGauge(features);
		patan.jte.applyFeatureGauge(patan.silhouettes, features, gauge);
		this._listFilter.refresh();
		this.backupSilhouettePuzzle();
	}

	// storage
	restoreSilhouettePuzzle(){
		var result = false;
		var bup = this._storage.getItem(this._storageKey);
		if( bup != null ){
			this._silhouettePuzzle = JSON.parse(bup, this._jte.jsonReviver.bind(this._jte));
			result = true;
		}
		return result;
	}

	backupSilhouettePuzzle(){
		this._storage.setItem(this._storageKey, JSON.stringify(this._silhouettePuzzle));
	}

	loadSilhouettePuzzle(xml, filename){
		var ret = {};
		try{
			ret = this._jte.expandJteXml(xml, this._silhouettePuzzle);
		}catch(e){
			return {'result': false, 'message': 'exception occured during jte expansion'};
		}
		if( ret.result ){
			this._silhouettePuzzle['source'] = filename;
			this._storage.setItem(this._storageKey, JSON.stringify (this._silhouettePuzzle));
			this._listFilter.reset();
		}
		return ret;
	}

	directSetSilhouettePuzzle(silhouettePuzzle, source){
		this._silhouettePuzzle = silhouettePuzzle;
		this._silhouettePuzzle['source'] = source;
		this._storage.setItem(this._storageKey, JSON.stringify (this._silhouettePuzzle));
		this._listFilter.reset();
	}

	removeStorage(){
		this._storage.removeItem(this._storageKey);
	}
}

//------ requirement
function checkRequirement(){
	if(
		window.File &&
		window.FileReader &&
		window.FileList &&
		window.Blob &&
		typeof Storage !== 'undefined'
	){
		return true;
	}
	return false;
}

//------ file
function fileSelectHandler(event) {
	loadJteFile(event.target.files);
}

function fileClearPath(event) {
	this.value = null;
}

function fileDragoverHandler(event) {
	var _event = ( event.originalEvent) ? event.originalEvent : event;
	event.stopPropagation();
	event.preventDefault();
	_event.dataTransfer.dropEffect = 'copy';
}

function fileDragleaveHandler(event) {
	event.stopPropagation();
	event.preventDefault();
}

function fileDropHandler(event) {
	var _event = ( event.originalEvent) ? event.originalEvent : event;
	event.stopPropagation();
	event.preventDefault();
	loadJteFile(_event.dataTransfer.files);
}

//------ application
var patan = null;
var currentIndex = null;
var deleteMode = false;
var logoCanvas = null;
var piecesetCanvas = null;
var listCanvas = null;
var workCanvas = null;

function onReady(){
	patan = new Patan();
	patan.localize.buildLanguageSelector({cn: 'language', fn: 'setLanguage',});
	patan.localize.resetLanguage();
	patan.colorTheme.applyTheme();

	// check requirement
	if(  !checkRequirement() ){
		$('#please_wait_layer').css('display','none');
		$('#env_error_layer').css('display','');
		return;
	}
}

function onLoad(){
	workCanvas = new WorkCanvas($('#workcanvas')[0]);
	patan.initialize(workCanvas.scope);

	logoCanvas = new LogoCanvas(patan, $('#logocanvas')[0]);
	piecesetCanvas = new PiecesetCanvas(patan, $('#piecesetcanvas')[0]);
	listCanvas = new ListCanvas(patan, $('#listcanvas')[0]);
	editCanvas = new EditCanvas(patan, $('#editcanvas')[0]);

	setSize();
	onSetSilhouettePuzzle();
	$('#sid_version').text(applicationVersion);
	$('#file').on('change', fileSelectHandler);
	$('#file').on('click', fileClearPath);
	$('#logocaption').on('dragover', fileDragoverHandler);
	$('#logocaption').on('dragleave', fileDragleaveHandler);
	$('#logocaption').on('drop', fileDropHandler);
	$('#please_wait_layer').css('display','none');
	$('#load_or_select_layer').css('display','');

	$(window).resize(resize);

	getTwsEntries();
}

function reset(){
	patan.removeStorage();
	location.reload();
}

function setSize(){
	var width = $(window).innerWidth() - 40 * 2;
	var height = $(window).innerHeight() - 40 * 2;
	patan.width = width;
	patan.height = height;
	$('#main_layer').css('width', width + 'px');
	$('#edit_layer').css('width', width + 'px');
	$('#listcanvas').css('width', width + 'px');
	$('#editcanvas').css('width', width + 'px');
}

function resize(){
	setSize();
	listCanvas.resize();
	editCanvas.resize();
};

function onSetSilhouettePuzzle(){
	logoCanvas.reset();
	clearDeleteMode();
	currentIndex = null;
	
	// set pieceset information
	piecesetCanvas.reset();
	listCanvas.reset(0);
	listCanvas.setOnChangePointIndex(onChangePointIndex);
	listCanvas.setOnClickSilhouette(onClickSilhouette);
	editCanvas.reset();

	$('#piecesetname').text(patan.pieceset.name);
	$('#piecesetcaption').text(patan.pieceset.silhouette.name);
	$('#editsilhouettename').val('');
	resetEditCanvasWrapBackgroundImage();
	resetSilhouetteNumText();
	resetPageStripHtml();
	resetSilhouetteTagsHtml();
}

function movePage(n){
	listCanvas.reset(n);
	resetPageStripHtml();
}

function loadJteFile(files){
	var file = files[0];
	if( !file ){
		return;
	}
	$('#load_or_select_layer').css('display','none');
	$('#please_wait_layer').css('display','');
	setTimeout(executeLoadJteFile, 150, file);
}

function executeLoadJteFile(file){
	var reader = new FileReader();
	reader.onload = function(event) {
		var ret = patan.loadSilhouettePuzzle(event.target.result, file.name);
		if( !ret.result ){
			window.alert(patan.localize.getCurrentLanguageString('sid_jtereaderror') +'\n(' +ret.message + ')');
			$('#please_wait_layer').css('display','none');
			$('#load_or_select_layer').css('display','');
			return;
		}
		// succeeded to load jte xml
		onSetSilhouettePuzzle();
		transPageFromLoadToMain();
	}
	reader.readAsText(file, 'UTF-8');
}

function resetEditCanvasWrapBackgroundImage(){
	var style = '';
	if( currentIndex != null ){
		style = 'url("' +
			workCanvas.getDataURL(
				patan.pieceset.svgPath,
				patan.pieceset.zoomRatio,
				patan.silhouettes[currentIndex].p,
				patan.colorTheme.getCurrentThemeColor('cid_light'),
				patan.colorTheme.getCurrentThemeColor('cid_light'),
				0.2
			).url + '")';
	}
	$('#editcanvaswrap').css('background-image', style);
}

function setLanguage(id){
	patan.localize.setLanguage(id);
	resetSilhouetteTagsHtml();
}

function toggleNightMode(){
	patan.colorTheme.toggleTheme();

	resetTwsEntriesHtml();
	resetEditCanvasWrapBackgroundImage();
	logoCanvas.refreshPiecesColor();
	piecesetCanvas.refreshPiecesColor();
	listCanvas.refreshPiecesColor();
	editCanvas.refreshPiecesColor();
}

function downloadJteXml(){
	var filename = patan.pieceset.name + '.jte';
	var jte = patan.jte.buildJteXml(patan.pieceset, patan.silhouettes);
	var blob = new Blob([jte], {'type': 'text/plain'});
	$('#download').attr('download', filename);
	$('#download').attr('href', window.URL.createObjectURL(blob));
	$('#download')[0].click();
}

function toggleListColoring(){
	listCanvas.toggleColoring();
}

function toggleEditColoring(){
	editCanvas.toggleColoring();
}

function scrollUpToTop(){
	$("html,body").animate({scrollTop:0}, 600, 'easeOutExpo');
}

function toggleTopMoreMenu(){
	$('#tid_topmoremenu').css('display', ($('#tid_topmoremenu').css('display') == 'none') ? '' : 'none');
	$('#topmoremenu').css('display', ($('#topmoremenu').css('display') == 'none') ? '' : 'none');
}

function toggleListMoreMenu(){
	$('#tid_listmoremenu').css('display', ($('#tid_listmoremenu').css('display') == 'none') ? '' : 'none');
	$('#listmoremenu').css('display', ($('#listmoremenu').css('display') == 'none') ? '' : 'none');
	clearDeleteMode();
}

function toggleEditMoreMenu(){
	$('#tid_editmoremenu').css('display', ($('#tid_editmoremenu').css('display') == 'none') ? '' : 'none');
	$('#editmoremenu').css('display', ($('#editmoremenu').css('display') == 'none') ? '' : 'none');
}

function sortByFeatures(){
	if( patan.silhouettes.length == 0 ){
		return;
	}
	$('#main_layer').css('display','none');
	$('#please_wait_layer').css('display','');
	setTimeout(executeSortingByFeatures, 150);
}

function executeSortingByFeatures(){
	patan.sortByFeatures();
	listCanvas.redraw();
	resetSilhouetteNumText();
	resetPageStripHtml();
	resetSilhouetteTagsHtml();
	$('#please_wait_layer').css('display','none');
	$('#main_layer').css('display','');
};

function toggleDeleteMode(){
	deleteMode = !deleteMode;
	$('#tid_deletemode').toggleClass('deletemode');
}

function clearDeleteMode(){
	if( deleteMode ){
		toggleDeleteMode();
	}
}

function onChangePointIndex(index){
	var title = (index != null) ? jointTags(patan.silhouettes[index]) : '';
	var prefix = deleteMode ? patan.localize.getCurrentLanguageString('sid_deletemode') : '';
	$('#listcanvas').attr('title', prefix + title);
	if( index != null ){
		if( !$('#listcanvas').hasClass('clickable') ){
			$('#listcanvas').addClass('clickable');
		}
	}else{
		$('#listcanvas').removeClass('clickable');
	}
}

function onClickSilhouette(index){
	if( index != null ){
		$('#listcanvas').removeClass('clickable');

		if(deleteMode){
			patan.removeSilhouette(index);
			listCanvas.redraw();
			if( currentIndex == index ){
				currentIndex = null;
				resetEditCanvasWrapBackgroundImage();
			}else{
				currentIndex = (currentIndex == null) ? null :
					((index < currentIndex) ? (currentIndex - 1) : currentIndex);
			}
			resetSilhouetteNumText();
			resetPageStripHtml();
			resetSilhouetteTagsHtml();
		}else{
			currentIndex = index;
			var silhouette = patan.silhouettes[index];
			resetEditCanvasWrapBackgroundImage();
			editCanvas.applyPieceInfo(silhouette.p);
			$('#editsilhouettename').val(jointTags(silhouette));
			$("html,body").animate({scrollTop:$('#editarea').offset().top}, 600, 'easeOutExpo');
		}
	}
}

function setSilhouette(isInsert){
	var silhouette = getEditSilhouette();
	if( isInsert || currentIndex == null ){
		currentIndex = 0;
		patan.insertSilhouette(currentIndex, silhouette);
	}else{
		patan.setSilhouette(currentIndex, silhouette);
	}
	editCanvas.fixInitPieceInfo();
	listCanvas.redraw();
	resetEditCanvasWrapBackgroundImage();
	resetSilhouetteNumText();
	resetPageStripHtml();
	resetSilhouetteTagsHtml();
	scrollUpToTop();
}

function transcribeSilhouette(){
	var silhouette = getEditSilhouette();
	patan.setPiecesetSilhouette(silhouette);
	logoCanvas.reset();
	piecesetCanvas.reset();
	$('#piecesetcaption').text(patan.pieceset.silhouette.name);
	scrollUpToTop();
}

function getEditSilhouette(){
	var silhouette = separateTags($('#editsilhouettename').val());
	silhouette.p = editCanvas.getCopiedPieceInfo();
	return silhouette;
}

function separateTags(nameAndTags){
	return patan.jte.separateTags(nameAndTags);
}

function jointTags(silhouette){
	return patan.jte.jointTags(silhouette.name, silhouette.tags);
}

function onClickTag(index){
	patan.setTagFilterByIndex(index);
	listCanvas.redraw();
	resetSilhouetteNumText();
	resetPageStripHtml();
	resetSilhouetteTagsHtml();
}

function resetSilhouetteNumText(){
	$('#silhouettenum').text(patan.silhouettesList.length + ' / ' + patan.silhouettes.length);
}

function resetPageStripHtml(){
	var pageStripHtmlParam = {'cn':'cn', 'pn': 'pn clickable', 'fn': 'movePage',};
	$('#pagestrip').html(listCanvas.getPageStripHtml(pageStripHtmlParam));
}

function resetSilhouetteTagsHtml(){
	var silhouetteTagsHtmlParam = {'cn':'txtlink', 'pn': '', 'fn': 'onClickTag',};
	$('#silhouettetags').html(patan.getSilhouetteTagsHtml(silhouetteTagsHtmlParam));
}

function imageDownload(){
	var url = getDataUrl();
	imageUrlDownload(url);
}

function imageDownload_wide(){
	var url = getDataUrl_wide();
	imageUrlDownload(url);
}

function imageUrlDownload(url){
	var filename = $('#editsilhouettename').val() + '.png';
	var mimeType = url.match(/(:)([a-z\/]+)(;)/)[2];
	var byteString = atob(url.split(',')[1]);
	var content = new Uint8Array(byteString.length);
	for( var i = 0; i < byteString.length; i++ ){
		content[i] = byteString.charCodeAt(i);
	}
	var blob = new Blob([content], {type: mimeType,});
	url = window.URL.createObjectURL(blob);
	$('#download').attr('download', filename);
	$('#download').attr('href', url);
	$('#download')[0].click();
}

function getDataUrl(){
	return editCanvas.getDataUrl();
}

function getDataUrl_wide(){
	var pieceInfo = editCanvas.getCopiedPieceInfo();
	var coloringKey = editCanvas.getCurrentColoringKey();
	var url =
	workCanvas.getDataURL_wide(
		patan.pieceset,
		'none',
		patan.colorTheme.getCurrentThemeColor('cid_fg'),
		pieceInfo,
		patan.colorTheme.getCurrentThemeColor('cid_fills')[coloringKey],
		patan.colorTheme.getCurrentThemeColor('cid_strokes')[coloringKey]
	).url;
	return url;
}

function showHint(){
	window.alert(patan.localize.getCurrentLanguageString('sid_hint'));
}

function prepareTweet(){
}

function transPageFromLoadToMain(){
	$('#please_wait_layer').css('display','none');
	$('#load_or_select_layer').css('display','none');
	$('#main_layer').css('display','');
}

function transPageFromMainToLoad(){
	$('#main_layer').css('display','none');
	$('#load_or_select_layer').css('display','');
	clearDeleteMode();
}
