// patan list canvas

class ListCanvas{
	constructor(patan, element){
		this._patan = patan;
		this._scope = new paper.PaperScope();
		this._scope.setup(element);

		this._layer = new this._scope.Layer();
		this._coloring = 0;
		this._perPage = 50;
		this._currentPage = 0;
		this._pointIndex = null;
		this._indexes = null;
		this._pointIndexAtMouseDown = null;
		this._onChangePointIndex = null;
		this._onClickSilhouette = null;

		this._scope.view.pause();
		this._scope.view.onMouseDown = this._onMouseDown.bind(this);
		this._scope.view.onMouseUp = this._onMouseUp.bind(this);
		this._scope.view.onMousEenter = this._onMouseEnter.bind(this);
		this._scope.view.onMouseMove = this._onMouseMove.bind(this);
		this._scope.view.onMouseLeave = this._onMouseLeave.bind(this);
	}

	reset(page){
		this._currentPage = page;
		this._redrawList();
	}

	resize(){
		this._redrawList();
	}

	redraw(){
		this._redrawList();
	}

	setOnChangePointIndex(callback){
		this._onChangePointIndex = callback;
	}

	setOnClickSilhouette(callback){
		this._onClickSilhouette = callback;
	}

	refreshPiecesColor(){
		this._setPiecesColor();
		this._scope.view.draw();
	}

	toggleColoring(){
		this._toggleColoring();
		this._setPiecesColor();
		this._scope.view.draw();
	}

	getPageStripHtml(classInfo){
		return this._getPageStripHtml(classInfo);
	}


//------ page strip
	_getPageStripHtml(classInfo){
		var num = this._patan.silhouettesList.length;
		var html = '';
		for( var i = 0; i < num / this._perPage; i++ ){
			var fCurr = (i == this._currentPage);
			html += '<span class="' +
				(fCurr ? classInfo['cn'] : classInfo['pn']) + '" ' +
				(fCurr ? '' : 'onclick="' + classInfo['fn'] + '(' + i + ');"') +
				'>' + (i+1) + '</span>';
		}
		return html;
	}

//------ appearance
	_toggleColoring(){
		this._coloring += 1;
		this._coloring %= this._patan.colorTheme.getCurrentThemeColorings()['variation'];
	}

	_setPiecesColor(){
		var colorings = this._patan.colorTheme.getCurrentThemeColorings();
		var fills = colorings['fills'][colorings['colorings'][this._coloring]];
		var strokes = colorings['strokes'][colorings['colorings'][this._coloring]];

		this._layer.children.forEach(function(group){
			for( var i = 0; i < group.children.length; i++ ){
				var fill = fills[i % fills.length];
				if( fill != 'none' ){
					group.children[i].fillColor = fill;
				}
				var stroke = strokes[i % strokes.length];
				group.children[i].strokeColor = stroke;
			}
		});
	}

	_redrawList(){
		var scope = this._scope;
		var layer = this._layer;

		var padding = 10;
		var lineHeight = 100;
		var lineOffset = 50;

		var pathData = this._patan.pieceset.svgPath;
		var zoomRatio = this._patan.pieceset.zoomRatio;
		var silhouettes = this._patan.silhouettesList;

		var maxPage = Math.max(0, Math.floor((silhouettes.length - 1) / this._perPage));
		if( this._currentPage > maxPage ){
			this._currentPage = maxPage;
		}

		var top = this._currentPage * this._perPage;
		var bottom = top + this._perPage;

		var width = this._patan.width
		var right = padding;
		var height = (silhouettes.length == 0) ? 0 : lineHeight;
		
		scope.view.pause();

		layer.removeChildren();

		this._indexes = [];
		for( var n = top; n < bottom && n < silhouettes.length; n++ ){
			var pieceInfo = silhouettes[n].p;
			var group = new scope.Group();
			for( var i = 0; i < pieceInfo.length; i++ ){
				var cpath = new scope.CompoundPath(pathData[pieceInfo[i].kind]);
				cpath.pivot = ([0,0]);
				cpath.matrix.a = pieceInfo[i].flip;
				cpath.rotate(pieceInfo[i].angle);
				cpath.translate(pieceInfo[i].position);
				group.addChild(cpath);
			}
			group.strokeCap = 'round';
			group.strokeJoin = 'bevel';
			group.strokeWidth = 4.0 * (0.2 * zoomRatio);
			group.scaling = 0.2 * zoomRatio;
			
			var w = group.bounds.width;
			if( right + w + padding < width ){
				group.position = [right + w/2, lineOffset];
			}else{
				right = padding;
				lineOffset += lineHeight;
				height += lineHeight;
				group.position = [right + w/2, lineOffset];
			}
			right += w + padding;
			
			this._indexes.push(this._patan.getRealSilhouetteIndex(n));
			layer.addChild(group);
		}
		this._setPiecesColor();

		scope.view.viewSize = new scope.Size(width, height);
		scope.view.draw();
		
		scope.view.play();
	}

//------ user interface
	// hit check
	_getPointIndex(p){
		for( var i = 0; i < this._layer.children.length; i++ ){
			if( this._layer.children[i].bounds.contains(p) ){
				return this._indexes[i];
			}
		}
		return null;
	}

	// for click
	_onMouseDown(event){
		this._pointIndexAtMouseDown = this._getPointIndex(event.point);
	}

	_onMouseUp(event){
		if( this._onClickSilhouette != null ){
			this._pointIndex = this._getPointIndex(event.point);
			if( this._pointIndex == this._pointIndexAtMouseDown ){
				this._onClickSilhouette(this._pointIndex);
			}
			this._pointIndexAtMouseDown = null;
		}
		return false;
	}

	// for move
	_onMouseEnter(event){
		this._pointIndexAtMouseDown = null;
		if( this._onChangePointIndex != null ){
			this._pointIndex = this._getPointIndex(event.point);
			this._onChangePointIndex(this._pointIndex);
		}
	}

	_onMouseMove(event){
		var index = this._getPointIndex(event.point);
		if( index != this._pointIndex ){
			this._pointIndex = index;
			if( this._onChangePointIndex != null ){
				this._onChangePointIndex(this._pointIndex);
			}
		}
	}

	_onMouseLeave(event){
		this._pointIndexAtMouseDown = null;
		if( this._onChangePointIndex != null ){
			this._pointIndex = null;
			this._onChangePointIndex(this._pointIndex);
		}
	}
}
